/****************************************************************************
 * @file     main.c
 * @version  V1.0
 * $Revision: 2 $
 * $Date: 17/12/21 10:04a $
 * @brief
 *           Implement transmit message in Normal mode.
 *           This sample code needs to work with CAN_NormalMode_Receive.
 * @note
 * Copyright (C) 2014~2017 Nuvoton Technology Corp. All rights reserved.
 *
 ******************************************************************************/
#include <stdio.h>
#include <string.h>
#include "M451Series.h"

#define V6M_AIRCR_VECTKEY_DATA            0x05FA0000UL
#define V6M_AIRCR_SYSRESETREQ             0x00000004UL
#define PLL_CLOCK                         48000000
#define CAN_BAUD_RATE                     250000
#define Master_ISP_ID                     0x800
#define Device0_ISP_ID                    0x802
#define CAN_ISP_DtatLength                0x08
#define CAN_RETRY_COUNTS                  0x8fffff
#define StartProgramAddress               0

/*---------------------------------------------------------------------------*/
/*  Function Declare                                                         */
/*---------------------------------------------------------------------------*/
extern char GetChar(void);

/* Declare a CAN message structure */
typedef struct
{
  uint32_t  Address;
  uint8_t   Data[8];
} STR_CANMSG_ISP;

STR_CANMSG_T rrMsg;
volatile uint8_t u8CAN_PackageFlag=0;
uint32_t Chip_EndAddress=0;

/*---------------------------------------------------------------------------------------------------------*/
/* ISR to handle CAN interrupt event                                                                       */
/*---------------------------------------------------------------------------------------------------------*/
void CAN_MsgInterrupt(CAN_T *tCAN, uint32_t u32IIDR)
{
    if(u32IIDR == 1)
    {
        CAN_Receive(tCAN, 0, &rrMsg);
        u8CAN_PackageFlag=1;
    }
}

/*---------------------------------------------------------------------------------------------------------*/
/* CAN0 interrupt handler                                                                                  */
/*---------------------------------------------------------------------------------------------------------*/
void CAN0_IRQHandler(void)
{
    uint32_t u8IIDRstatus;

    u8IIDRstatus = CAN0->IIDR;

    if(u8IIDRstatus == 0x00008000)        /* Check Status Interrupt Flag (Error status Int and Status change Int) */
    {
        /**************************/
        /* Status Change interrupt*/
        /**************************/
        if(CAN0->STATUS & CAN_STATUS_RXOK_Msk)
        {
            CAN0->STATUS &= ~CAN_STATUS_RXOK_Msk;   /* Clear RxOK status*/
        }

        if(CAN0->STATUS & CAN_STATUS_TXOK_Msk)
        {
            CAN0->STATUS &= ~CAN_STATUS_TXOK_Msk;    /* Clear TxOK status*/
        }
    }
    else if(u8IIDRstatus!=0)
    {
      CAN_MsgInterrupt(CAN0, u8IIDRstatus);  

      CAN_CLR_INT_PENDING_BIT(CAN0, (u8IIDRstatus - 1)); /* Clear Interrupt Pending */
    }
}

void SYS_Init(void)
{
    /*---------------------------------------------------------------------------------------------------------*/
    /* Init System Clock                                                                                       */
    /*---------------------------------------------------------------------------------------------------------*/

    /* Enable HIRC clock (Internal RC 22.1184MHz) */
    CLK->PWRCTL |= CLK_PWRCTL_HIRCEN_Msk;

    /* Waiting for HIRC clock ready */
    while(!(CLK->STATUS & CLK_STATUS_HIRCSTB_Msk));

    /* Select HCLK clock source as HIRC and and HCLK clock divider as 1 */
    CLK->CLKSEL0 &= ~CLK_CLKSEL0_HCLKSEL_Msk;
    CLK->CLKSEL0 |= CLK_CLKSEL0_HCLKSEL_HIRC;
    CLK->CLKDIV0 &= ~CLK_CLKDIV0_HCLKDIV_Msk;
    CLK->CLKDIV0 |= CLK_CLKDIV0_HCLK(1);

    /* Set PLL to Power-down mode and PLLSTB bit in CLK_STATUS register will be cleared by hardware.*/
    CLK->PLLCTL |= CLK_PLLCTL_PD_Msk;

    /* Enable HXT clock (external XTAL 12MHz) */
    CLK->PWRCTL |= CLK_PWRCTL_HXTEN_Msk;

    /* Waiting for HXT clock ready */
    while(!(CLK->STATUS & CLK_STATUS_HXTSTB_Msk));

    /* Set core clock as PLL_CLOCK from PLL */
    CLK->PLLCTL = (CLK_PLLCTL_PLLSRC_HXT  | CLK_PLLCTL_NR(2) | CLK_PLLCTL_NF(32) | CLK_PLLCTL_NO_4);
    while(!(CLK->STATUS & CLK_STATUS_PLLSTB_Msk));
    CLK->CLKSEL0 &= (~CLK_CLKSEL0_HCLKSEL_Msk);
    CLK->CLKSEL0 |= CLK_CLKSEL0_HCLKSEL_PLL;

    /* Update System Core Clock */
    PllClock        = PLL_CLOCK;            // PLL
    SystemCoreClock = PLL_CLOCK / 1;        // HCLK
    CyclesPerUs     = PLL_CLOCK / 1000000;  // For CLK_SysTickDelay()    
    
    /* Enable UART module clock */
    CLK->APBCLK0 |= CLK_APBCLK0_UART3CKEN_Msk;    
    
    /* Select UART module clock source as HXT and UART module clock divider as 1 */
    CLK->CLKSEL1 &= ~CLK_CLKSEL1_UARTSEL_Msk;
    CLK->CLKSEL1 |= CLK_CLKSEL1_UARTSEL_HXT;
    CLK->CLKDIV0 &= ~CLK_CLKDIV0_UARTDIV_Msk;
    CLK->CLKDIV0 |= CLK_CLKDIV0_UART(1);    
    
    /* Enable CAN0 module clock */
    CLK->APBCLK0 |= CLK_APBCLK0_CAN0CKEN_Msk;    
}

/*----------------------------------------------------------------------------*/
/*  Tx Msg by Normal Mode Function (With Message RAM)                    */
/*----------------------------------------------------------------------------*/
void CAN_Package_Tx(CAN_T *tCAN, STR_CANMSG_ISP *CANBuffer)
{
    STR_CANMSG_T tMsg;
    uint32_t APROM_Data=0;
    /* Send a 11-bit Standard Identifier message */
    tMsg.FrameType = CAN_DATA_FRAME;
    tMsg.IdType    = CAN_STD_ID;
    tMsg.Id        = Master_ISP_ID;
    tMsg.DLC       = CAN_ISP_DtatLength;
    memcpy(&tMsg.Data, &CANBuffer->Address, 4);    //address
    APROM_Data = FMC_Read(CANBuffer->Address);
    memcpy(&tMsg.Data[4], &APROM_Data, 4);         //data
    if(CAN_Transmit(tCAN, MSG(5), &tMsg) == FALSE)  // Configure Msg RAM and send the Msg in the RAM
    {
        return;
    }
}

uint8_t can_setRxMsg(CAN_T *tCAN)
{
    if(CAN_SetRxMsg(tCAN, MSG(0), CAN_STD_ID, Device0_ISP_ID) == FALSE)
    {
        return FALSE;
    }
    return TRUE;
}

/*----------------------------------------------------------------------------*/
/*  Rx Msg by Normal Mode Function (With Message RAM)                    */
/*----------------------------------------------------------------------------*/
uint8_t CAN_ISP_Update(CAN_T *tCAN)
{
    STR_CANMSG_ISP CAN_buffer;
    uint32_t u32TimeOutCount = 0;
    uint8_t ErrorCnt=0, TimeOut_flag=1, CompareFail_flag=1;
    memset(&CAN_buffer, 0, sizeof(CAN_buffer));
    CAN_buffer.Address = StartProgramAddress;

    while(1)
    {
        if(ErrorCnt++ == 0xFF) return FALSE;             //ErrorCnt = 0xff CAN bus fail
        CAN_Package_Tx(CAN0, &CAN_buffer);               //send CAN ISP Package
    
        u32TimeOutCount = 0;
        u8CAN_PackageFlag=0;
        TimeOut_flag=1;
        while(u8CAN_PackageFlag == 0)                    //Wait CAN ISP Package ACK
        {
            if(++u32TimeOutCount >= CAN_RETRY_COUNTS)
            {
                TimeOut_flag=0;
                break;
            }
        }
        if(TimeOut_flag)
        {
            CompareFail_flag=1;
            if(((long)CAN_buffer.Address) != (*(long*)rrMsg.Data)) CompareFail_flag=0;   //check CAN ISP Package
            if(CompareFail_flag)
            {
                if(CAN_buffer.Address == Chip_EndAddress)  return TRUE;                                             //check CAN ISP EndCode
                ErrorCnt=0;
                CAN_buffer.Address += 4;                                                                                 //for next address            
            }
            else
            {
                if(CAN_buffer.Address < FMC_FLASH_PAGE_SIZE) CAN_buffer.Address = StartProgramAddress;
                else CAN_buffer.Address = (CAN_buffer.Address / FMC_FLASH_PAGE_SIZE) * FMC_FLASH_PAGE_SIZE;
            }
        }
    }
}

void can_init(void)
{
    /* Set PA multi-function pins for CAN0 RXD and TXD */
    SYS->GPA_MFPH &= ~(SYS_GPA_MFPH_PA12MFP_Msk | SYS_GPA_MFPH_PA13MFP_Msk);
    SYS->GPA_MFPH |= (SYS_GPA_MFPH_PA12MFP_CAN0_TXD | SYS_GPA_MFPH_PA13MFP_CAN0_RXD);
    
    /* Set PA.0 and PA.1 for GPIO to control CAN transciver */
    SYS->GPA_MFPL &= ~(SYS_GPA_MFPL_PA0MFP_Msk | SYS_GPA_MFPL_PA1MFP_Msk);
    SYS->GPA_MFPL |= (SYS_GPA_MFPL_PA0MFP_GPIO | SYS_GPA_MFPL_PA1MFP_GPIO);
    
    /* Set CAN transceiver to high speed mode */
    GPIO_SetMode(PA, (BIT0 | BIT1), GPIO_MODE_OUTPUT);
    PA0 = 0;
    PA1 = 0;
    
    CAN_Open(CAN0, CAN_BAUD_RATE, CAN_NORMAL_MODE);

    CAN_EnableInt(CAN0, (CAN_CON_IE_Msk | CAN_CON_SIE_Msk | CAN_CON_EIE_Msk));
    NVIC_SetPriority(CAN0_IRQn, (1 <<__NVIC_PRIO_BITS) - 2);
    NVIC_EnableIRQ(CAN0_IRQn);
        
    /* Set CAN reveive message */
    can_setRxMsg(CAN0);
}

/*---------------------------------------------------------------------------------------------------------*/
/* MAIN function                                                                                           */
/*---------------------------------------------------------------------------------------------------------*/
int main(void)
{
    /* Unlock protected registers */
    SYS_UnlockReg();
    /* Init System, IP clock and multi-function I/O */
    SYS_Init();
    /* Init CAN port */
    can_init();
    /* Enable FMC ISP function */
    FMC_Open();
    if((FMC_Read(0x20000) == 0) && ((FMC->ISPCTL & FMC_ISPCTL_ISPFF_Msk) == FMC_ISPCTL_ISPFF_Msk)) Chip_EndAddress = (128 * 1024)-4;
    else Chip_EndAddress = (256 * 1024)-4;
    /* stat update program */
    while(1)
    {
        if(CAN_ISP_Update(CAN0)>0)
        {
            goto _APROM;
        }
        while(1);
    }
_APROM:
    outpw(&SYS->RSTSTS, 3);//clear bit
    outpw(&FMC->ISPCTL, inpw(&FMC->ISPCTL) & 0xFFFFFFFC);
    outpw(&SCB->AIRCR, (V6M_AIRCR_VECTKEY_DATA | V6M_AIRCR_SYSRESETREQ));
    /* Trap the CPU */
    while(1);
}



